/*-*-c++-*-*/
/******************************************************************************
 **                                                                          **
 ** Copyright (c) 2000 Massachusetts Institute of Technology, Cambridge, MA. **
 ** All rights reserved.                                                     **
 **                                                                          **
 ******************************************************************************
  ==========================================================================
  ============================= File Info ==================================

  Author: Zhenhai Zhu
  
  Description:

  Resources:

  See also:

  ==========================================================================
*/

const static char cvsid[] = "$Id: grid.cc,v 1.5 2002/12/03 14:49:33 zhzhu Exp $";

#include "grid.h"
#include "gridElement.h"
#include "stencil.h"

using namespace pfft;
using namespace std;

/**********************************************************************
 * Grid --
 **********************************************************************/
Grid::Grid (
	    GridElement* srcGridElement,
	    GridElement* evalGridElement,
	    Stencil* projectStencil,
	    Stencil* interpStencil,
	    Stencil* directStencil,
	    double maxGridSize)
  : srcGridElementPtr(srcGridElement), evalGridElementPtr(evalGridElement), 
  projectStencilPtr(projectStencil), interpStencilPtr(interpStencil),
  directStencilPtr(directStencil), maxGridSize_(maxGridSize)
{
  averageElementSize_ = (srcGridElementPtr->averageElementSize() + 
			 evalGridElementPtr->averageElementSize() ) / 2.;
  initialGridParameter();
}

/**********************************************************************
 * initialGridParameter --
 **********************************************************************/
inline void
Grid::initialGridParameter (
			    void)
{
  numPointX_ = numPointY_ = numPointZ_ = 4;
  bakNumPointX_ = bakNumPointY_ = bakNumPointZ_ = numPointX_;
  prevGridVolume = currentGridVolume = DBL_MAX;
}

/**********************************************************************
 * setupGridSize --
 **********************************************************************/
void 
Grid::setupGridSize (
		     void)
{
  // find the range of problem domain, which includes src and eval gridElements
  double xmin = min(srcGridElementPtr->xmin(), evalGridElementPtr->xmin());
  double xmax = max(srcGridElementPtr->xmax(), evalGridElementPtr->xmax());
  double ymin = min(srcGridElementPtr->ymin(), evalGridElementPtr->ymin());
  double ymax = max(srcGridElementPtr->ymax(), evalGridElementPtr->ymax());
  double zmin = min(srcGridElementPtr->zmin(), evalGridElementPtr->zmin());
  double zmax = max(srcGridElementPtr->zmax(), evalGridElementPtr->zmax());

  // Add a (very) little slack to the grid, so that all elements are fully inside the
  // grid, i.e. no part of any element is on the border of the grid.
  xmin -= DBL_EPSILON*abs(xmin);
  xmax += DBL_EPSILON*abs(xmax);
  ymin -= DBL_EPSILON*abs(ymin);
  ymax += DBL_EPSILON*abs(ymax);
  zmin -= DBL_EPSILON*abs(zmin);
  zmax += DBL_EPSILON*abs(zmax);

  // Add extra layers to the grids embedded the elements
  const double extraLayer = 
    max(projectStencilPtr->size(), interpStencilPtr->size()) - 0.5;

  // extralayers are counted in the numPointX, but xmax and xmin do not include them
  // It is possible that the number of extra layers is bigger them initial numPoint along 
  // X, Y or Z. Hence I use a guard here.
  double dx = (xmax - xmin) / max(1., (numPointX_ - 2*extraLayer - 1));
  double dy = (ymax - ymin) / max(1., (numPointY_ - 2*extraLayer - 1));
  double dz = (zmax - zmin) / max(1., (numPointZ_ - 2*extraLayer - 1));
  gridSize_ = max(max(dx, dy), dz);

  // Find out along which direction the elements fit tightly with grids
  // This direction will be the potential direction along which the grid number
  // should be increased if necessary
  if (dx == gridSize_) tightFit_ = 0;
  else if (dy == gridSize_) tightFit_ = 1;
  else  tightFit_ = 2;

  // for some reason, the uniform grid is used here
  dx_ = gridSize_;
  dy_ = gridSize_;
  dz_ = gridSize_;

  // one extra layer to each side
  xmin_ = xmin - extraLayer * dx_; 
  ymin_ = ymin - extraLayer * dy_;
  zmin_ = zmin - extraLayer * dz_;
  xmax_ = xmin_ + (numPointX_ - 1) * dx_;
  ymax_ = ymin_ + (numPointY_ - 1) * dy_;
  zmax_ = zmin_ + (numPointZ_ - 1) * dz_;

  x_.resize(numPointX_);
  for (size_t i = 0; i < numPointX_; i++)
    x_[i] = xmin_ + i * dx_;

  y_.resize(numPointY_);
  for (size_t i = 0; i < numPointY_; i++)
    y_[i] = ymin_ + i * dy_;

  z_.resize(numPointZ_);
  for (size_t i = 0; i < numPointZ_; i++)
    z_[i] = zmin_ + i * dz_;

  internalGridReport();

}

/**********************************************************************
 * checkGridSize --
 **********************************************************************/
Grid::GridStatus 
Grid::checkGridSize (
		     void)
{
  // some hard-coded numbers used here
  const double maxGridElementRatio = 10;
  const double minGridElementRatio = 0.5;
  const double minInteractVolumeRatio = 0.5;
  const size_t maxNumElementToOneGrid = 110;

  // Make sure the gridSize is no larger than a user-specified upper bound
  // If user does not specify, this constrain has no effect
  if (gridSize_ > maxGridSize_) {
    debugReport("Grid too coarse: grid size is larger than user specified threshold");
    return TOO_COARSE;
  }

  // if a grid has too many elements mapped to it, it means the grid is too coarse
  if ( (srcGridElementPtr->maxNumElementMappedToOneGrid() > maxNumElementToOneGrid) ||
       (evalGridElementPtr->maxNumElementMappedToOneGrid() > maxNumElementToOneGrid) ) {
    debugReport("Max number of elements mapped to one grid := ", 
		srcGridElementPtr->maxNumElementMappedToOneGrid());
    debugReport("Grid too coarse: too many elements associated with one grid point.");
    return TOO_COARSE;
  }

  // make sure gridSize and elementSize are reasonably proportional
  if (gridSize_ < averageElementSize_* minGridElementRatio) {
    debugReport("Grid too fine: grid size is too small comparing to min element size");
    return TOO_FINE;
  }
  if (gridSize_ > averageElementSize_* maxGridElementRatio) {
    debugReport("Grid too coarse: grid size is too large comparing to max element size");
    return TOO_COARSE; 
  }

  // Make sure the domain occupied by current grid is no larger than previous one
  prevGridVolume = currentGridVolume;
  currentGridVolume = (xmax_-xmin_) * (ymax_-ymin_) * (zmax_-zmin_);
  if (currentGridVolume > prevGridVolume) {
    debugReport("Grid too coarse: current grid volume is larger than previous one");
    return TOO_COARSE;
  }

  // make sure the interactive range defined by directStencil is not too small
  // comparing to the interactive range defined by element size.
  // This is one more constrain on gridSize vs. elementSize.
  double stencilVolume = directStencilPtr->numPoint() * dx_ * dy_ * dz_;
  double averageRadius = averageElementSize_ / 2.;
  double interactVolumeRadius = averageRadius * (1. + separationFactor());
  double interactVolume = 4./3*PAI*pow(interactVolumeRadius, 3);
  if (stencilVolume < interactVolume * minInteractVolumeRatio) {
    debugReport("Grid too fine: interactive volume defined by stencil is too small");
    return TOO_FINE;
  }

  debugReport("Grid size OK");
  return SIZE_OK;
}

/**********************************************************************
 * doubleNumPoint --
 **********************************************************************/
void
Grid::doubleNumPoint (
		      void)
{
  if (tightFit_ == 0) 
    numPointX_ *= 2;
  else if (tightFit_ == 1) 
    numPointY_ *= 2;
  else  
    numPointZ_ *= 2;
}

/**********************************************************************
 * transfer1DIndexTo3D --
 **********************************************************************/
GridIndex
Grid::transfer1DIndexTo3D (
			   int index1D) const
{
  int x = static_cast<int>( floor(index1D / static_cast<double>(2*numPointZ_) / (2*numPointY_)) );
  index1D -= x * (2*numPointZ_) * (2*numPointY_);
  int y = static_cast<int>( floor(index1D / static_cast<double>((2*numPointZ_))) );
  int z = index1D - y * (2*numPointZ_);

  return GridIndex(x, y, z);
}

/**********************************************************************
 * gridPointDistance --
 **********************************************************************/
double
Grid::distance(
	       const int ix1, 
	       const int iy1, 
	       const int iz1,
	       const int ix2, 
	       const int iy2, 
	       const int iz2) const
{
  double x = dx_ * (ix2 - ix1);
  double y = dy_ * (iy2 - iy1);
  double z = dz_ * (iz2 - iz1);
  return sqrt(x*x + y*y + z*z);
}

/**********************************************************************
 * debugReport --
 **********************************************************************/
inline void 
Grid::debugReport (
		   string message)
{
#ifdef DEBUG_GRID
    cout << message << endl;
#endif
}

/**********************************************************************
 * debugReport --
 **********************************************************************/
template <class T>
inline void 
Grid::debugReport (
		   const string message,
		   const T number)
{
#ifdef DEBUG_GRID
  cout << message << " " << number << endl;
#endif
}

/**********************************************************************
 * internalGridReport --
 **********************************************************************/
void
Grid::internalGridReport (
			  void) const
{
#ifdef DEBUG_GRID
  cout << "number of grids := (" 
       << numPointX_ << "," 
       << numPointY_ << "," 
       << numPointZ_ << ")"
       << "  grid size := " 
       << gridSize_
       << endl;
#endif
}

