/*-*-c++-*-*/
/******************************************************************************
 **                                                                          **
 ** Copyright (c) 2000 Massachusetts Institute of Technology, Cambridge, MA. **
 ** All rights reserved.                                                     **
 **                                                                          **
 ******************************************************************************
  ==========================================================================
  ============================= File Info ==================================

  Author: Zhenhai Zhu
  
  Description:

  Resources:

  See also:

  ==========================================================================
*/

const static char cvsid[] = "$Id: stencil.cc,v 1.8 2002/12/10 23:39:56 zhzhu Exp $";

#include "dense.h" // for pseudo_inverse()
#include <cstdlib> // for new and delete
#include <stdexcept> //for exception handling
#include "stencil.h"
#include "grid.h"
#include "vector3D.h"

using namespace std;
using namespace pfft;
using namespace TNT;

/**********************************************************************
 * stencil --
 **********************************************************************/
Stencil::Stencil(
		 const StencilShape shape, 
		 const StencilType type,
		 const int size, 
		 const BasisType basisType,
		 const double step) 
  : stencilShape_(shape), stencilType_(type), size_(size), 
  step_(step), basisType_(basisType)
{
  setupStencilPoint();
  setupPolynomialInterp();
}

/**********************************************************************
 * setup --
 **********************************************************************/
void
Stencil::setup(
	       const double step) 
{
  step_ = step;
  setupPolynomialInterp();
}

/**********************************************************************
 * setupStencilPoint --
 **********************************************************************/
void
Stencil::setupStencilPoint (
			    void)
{
  if (stencilShape_ == CUBE_STENCIL) {
    findCubeStencilPoints();
  } else if (stencilShape_ == SPHERE_STENCIL) {
    findSphereStencilPoints();
  } else {
    throw domain_error("unknown stencil type!");
  }
  totalNumPoint_ = points_.size();
}

/**********************************************************************
 * findSphereStencilPoints --
 * Note:  the loop sequence in this function should be consistent with
 * that in findGlobalIndexOfSphereStencilPoint
 **********************************************************************/
void
Stencil::findSphereStencilPoints (
				  void)
{
  int min = -size_;
  int max = size_;

  int NN = 2*size_ + 1;
  local1DIndexArrayForSphereStencil_ = 
    vector<TNT::Matrix<int> >(NN, Matrix<int>(NN, NN));
  
  points_.clear();

  int currentLocal1DIndex = 0;
  for (int x=min; x<=max; x++) {
    for (int y=min; y<=max; y++) {
      for (int z=min; z<=max; z++) {
	GridIndex grid(x, y, z);
	if (grid.length2() <= size_*size_) {
	  points_.push_back(grid);
	  local1DIndexArrayForSphereStencil_[x+size_][y+size_][z+size_]
	    = currentLocal1DIndex ++;
	} else {
	  local1DIndexArrayForSphereStencil_[x+size_][y+size_][z+size_]
	    = -1;
	}
      }
    }
  }
}

/**********************************************************************
 * findCubeStencilPoints --
 * Note:  the loop sequence in this function should be consistent with
 * that in findGlobalIndexOfCubeStencilPoint
 **********************************************************************/
void
Stencil::findCubeStencilPoints (
				void)
{
  int min = -size_;
  int max = size_;

  points_.clear();
  for (int x=min; x<=max; x++) {
    for (int y=min; y<=max; y++) {
      for (int z=min; z<=max; z++) {
	GridIndex grid(x, y, z);
	points_.push_back(grid);
      }
    }
  }
}

/**********************************************************************
 * isInsideStencil --
 **********************************************************************/
bool
Stencil::isInsideStencil (
			  const int x, 
			  const int y,
			  const int z) const
{
  if (stencilShape_ == CUBE_STENCIL) {
    return (abs(x) <= size_) && (abs(y) <= size_) && (abs(z) <= size_);
  } else if (stencilShape_ == SPHERE_STENCIL) {
    GridIndex grid(x, y, z);
    return grid.length2() <= size_*size_;
  } else {
    throw domain_error("unknown stencil type!");
  }
}

/**********************************************************************
 * isInsideStencil --
 **********************************************************************/
bool 
Stencil::isInsideStencil (
			  const GridIndex& index3D) const
{
  if (stencilShape_ == CUBE_STENCIL) {
    return (abs(index3D.x()) <= size_) && 
      (abs(index3D.y()) <= size_) && 
      (abs(index3D.z()) <= size_);
  } else if (stencilShape_ == SPHERE_STENCIL) {
    return index3D.length() <= size_;
  } else {
    throw domain_error("unknown stencil type!");
  }
}

/**********************************************************************
 * findGlobalIndexOfStencilPoint --
 **********************************************************************/
void
Stencil::findGlobalIndexOfStencilPoint(
				       const int ihx, 
				       const int ihy, 
				       const int ihz,
				       const Grid& grid,
				       vector<int>& globalIndexList) const
{
  if (stencilShape_ == CUBE_STENCIL) {
    findGlobalIndexOfCubeStencilPoint(ihx, ihy, ihz, grid, globalIndexList);
  } else if (stencilShape_ == SPHERE_STENCIL) {
    findGlobalIndexOfSphereStencilPoint(ihx, ihy, ihz, grid, globalIndexList);
  }
}

/**********************************************************************
 * findGlobalIndexOfCubeStencilPoint --
 **********************************************************************/
inline void
Stencil::findGlobalIndexOfCubeStencilPoint(
					   const int ihx, 
					   const int ihy, 
					   const int ihz,
					   const Grid& grid,
					   vector<int>& globalIndexList) const
{
  // remove any points outside the grid range
  int xmin = max(ihx - size_, 0);
  int xmax = min(ihx + size_, static_cast<int>(grid.numPointX()-1));
  int ymin = max(ihy - size_, 0);
  int ymax = min(ihy + size_, static_cast<int>(grid.numPointY()-1));
  int zmin = max(ihz - size_, 0);
  int zmax = min(ihz + size_, static_cast<int>(grid.numPointZ()-1));

  for (int igx = xmin; igx <= xmax; igx++) {
    for (int igy = ymin; igy <= ymax; igy++) {
      for (int igz = zmin; igz <= zmax; igz++) {
	globalIndexList.push_back(grid.transfer3DIndexTo1D(igx, igy, igz));
      }
    }
  }
}

/**********************************************************************
 * findGlobalIndexOfSphereStencilPoint --
 **********************************************************************/
inline void
Stencil::findGlobalIndexOfSphereStencilPoint(
					     const int ihx, 
					     const int ihy, 
					     const int ihz,
					     const Grid& grid,
					     vector<int>& globalIndexList) const
{
  // remove any points outside the grid range
  int xmin = max(ihx - size_, 0);
  int xmax = min(ihx + size_, static_cast<int>(grid.numPointX()-1));
  int ymin = max(ihy - size_, 0);
  int ymax = min(ihy + size_, static_cast<int>(grid.numPointY()-1));
  int zmin = max(ihz - size_, 0);
  int zmax = min(ihz + size_, static_cast<int>(grid.numPointZ()-1));

  for (int igx = xmin; igx <= xmax; igx++) {
    for (int igy = ymin; igy <= ymax; igy++) {
      for (int igz = zmin; igz <= zmax; igz++) {
	if(isInsideStencil(ihx-igx, ihy-igy, ihz-igz)) {
	  globalIndexList.push_back(grid.transfer3DIndexTo1D(igx, igy, igz));
	}
      }
    }
  }
}


/**********************************************************************
 * findGlobal3DIndexOfStencilPoint --
 **********************************************************************/
void
Stencil::findGlobal3DIndexOfStencilPoint(
					 const GridIndex& hostPoint, 
					 const Grid& grid,
					 vector<GridIndex>& globalIndexList) const
{
  if (stencilShape_ == CUBE_STENCIL) {
    findGlobal3DIndexOfCubeStencilPoint(hostPoint, grid, globalIndexList);
  } else if (stencilShape_ == SPHERE_STENCIL) {
    findGlobal3DIndexOfSphereStencilPoint(hostPoint, grid, globalIndexList);
  }
}

/**********************************************************************
 * findGlobal3DIndexOfCubeStencilPoint --
 **********************************************************************/
inline void
Stencil::findGlobal3DIndexOfCubeStencilPoint(
					     const GridIndex& hostPoint, 
					     const Grid& grid,
					     vector<GridIndex>& globalIndexList) const
{
  // remove any points outside the grid range
  int xmin = max(hostPoint.x() - size_, 0);
  int xmax = min(hostPoint.x() + size_, static_cast<int>(grid.numPointX()-1));
  int ymin = max(hostPoint.y() - size_, 0);
  int ymax = min(hostPoint.y() + size_, static_cast<int>(grid.numPointY()-1));
  int zmin = max(hostPoint.z() - size_, 0);
  int zmax = min(hostPoint.z() + size_, static_cast<int>(grid.numPointZ()-1));

  for (int igx = xmin; igx <= xmax; igx++) {
    for (int igy = ymin; igy <= ymax; igy++) {
      for (int igz = zmin; igz <= zmax; igz++) {
	globalIndexList.push_back(GridIndex(igx, igy, igz));
      }
    }
  }
}

/**********************************************************************
 * findGlobal3DIndexOfSphereStencilPoint --
 **********************************************************************/
inline void
Stencil::findGlobal3DIndexOfSphereStencilPoint(
					       const GridIndex& hostPoint, 
					       const Grid& grid,
					       vector<GridIndex>& globalIndexList) const
{
  // remove any points outside the grid range
  int xmin = max(hostPoint.x() - size_, 0);
  int xmax = min(hostPoint.x() + size_, static_cast<int>(grid.numPointX()-1));
  int ymin = max(hostPoint.y() - size_, 0);
  int ymax = min(hostPoint.y() + size_, static_cast<int>(grid.numPointY()-1));
  int zmin = max(hostPoint.z() - size_, 0);
  int zmax = min(hostPoint.z() + size_, static_cast<int>(grid.numPointZ()-1));

  for (int igx = xmin; igx <= xmax; igx++) {
    for (int igy = ymin; igy <= ymax; igy++) {
      for (int igz = zmin; igz <= zmax; igz++) {
	if(isInsideStencil(hostPoint.x()-igx, hostPoint.y()-igy, hostPoint.z()-igz)) {
	  globalIndexList.push_back(GridIndex(igx, igy, igz));
	}
      }
    }
  }
}

/**********************************************************************
 * setupComplete --
 **********************************************************************/
bool
Stencil::setupComplete (
			void) const
{
  if ( (stencilType_ == PROJECT_STENCIL) || (stencilType_ == INTERP_STENCIL) ) {
    return (! basisValue_.empty()) && (! weight_.empty());
  } else {
    return (! points_.empty());
  }
}


/**********************************************************************
 * setupPolynomialInterp --
 **********************************************************************/
void
Stencil::setupPolynomialInterp (
				void)
{
  if ( (stencilType_ != PROJECT_STENCIL) && (stencilType_ != INTERP_STENCIL) ) {
    return;
  }

  try {
    findBasisFuncTypeAndNumber();

    // I use step_=0 as a sign that no interp info is to be setup now becasue
    // stepSize has not be decided. It will be decided later. And since little 
    // work is done up to this point, it does not hurt to call this function later.  
    // The above function has to be called before return because the 
    // parameter "numBasis_" is needed to estimate memory usage in choosing 
    // optimal gridsize.
    if (step_ == 0.) {
      return;
    }

    setupEquaScale();
    setupPolyOrder();
    setupBasisValue();
    setupWeight();
  }
  catch (const std::domain_error error) {
    cerr << error.what() << endl;
    cerr << "there must be a bug here, exit ..." << endl;
    exit(1);
  }
}

/**********************************************************************
 * findBasisFuncTypeAndNumber --
 **********************************************************************/
void
Stencil::findBasisFuncTypeAndNumber (
				     void)
{
  equaScaleFlag_ = false;

  switch (basisType_) {
  case CUBED_POLY:
    numBasis_ = totalNumPoint_;
    maxMonoOrder_ = 
      static_cast<size_t>(floor(pow((double)totalNumPoint_, 1./3)+0.5)) - 1;
    break;
  case SPECIAL_POLY:
    if (totalNumPoint_ == 7) {
      numBasis_ = 7;
      maxMonoOrder_ = 2;
    } else {
      throw domain_error("stencil.cc:findBasisFuncTypeAndNumber-- unknown number of stencil points for SPECIAL_POLY");
    }
    break;
  case CONSISTENT_POLY:
    maxMonoOrder_ = size_ - 1;
    numBasis_ = (maxMonoOrder_ + 1) * (maxMonoOrder_ + 2) * (maxMonoOrder_ + 3) / 6;
    equaScaleFlag_ = true;
    break;
  default:
    throw domain_error("stencil.cc:findBasisFuncTypeAndNumber-- unknown number of stencil points");
    break;
  }
}

/**********************************************************************
 * setupEquaScale --
 **********************************************************************/
void
Stencil::setupEquaScale (
			 void)
{
  equaScale_ = vector<double>(totalNumPoint_, 1.);
  if (equaScaleFlag_) {
    int pointIndex = 0;
    for (vector<GridIndex>::const_iterator it = points_.begin(); it != points_.end(); ++it, pointIndex++) {
      switch (it->length2()) {
      case 0:
	equaScale_[pointIndex] = 1.0;
	break;
      case 1:
	equaScale_[pointIndex] = 0.39;
	break;
      case 2:
      case 3:
	equaScale_[pointIndex] = 0.22;
	break;
      case 4:
	equaScale_[pointIndex] = 0.02;
	break;
      default:
	throw domain_error("stencil.cc:setupEquaScale--  grid point index error!!");
	break;
      }
    }
  }
}

/**********************************************************************
 * setupPolyOrder --
 **********************************************************************/
void
Stencil::setupPolyOrder (
			 void)
{
  xMonoOrder_.reserve(numBasis_);
  yMonoOrder_.reserve(numBasis_);
  zMonoOrder_.reserve(numBasis_);

  switch (basisType_) {
  case CUBED_POLY:
    maxPolyOrder_ = 3*maxMonoOrder_;
    for (int x = 0; x <= maxMonoOrder_; x++) 
      for (int y = 0; y <= maxMonoOrder_; y++) 
	for (int z = 0; z <= maxMonoOrder_; z++) {
	  xMonoOrder_.push_back(x);
	  yMonoOrder_.push_back(y);
	  zMonoOrder_.push_back(z);
	}
    break;
  case CONSISTENT_POLY:
    maxPolyOrder_ = maxMonoOrder_;
    for (int x = 0; x <= maxMonoOrder_; x++) 
      for (int y = 0; y <= maxMonoOrder_ - x; y++) 
	for (int z = 0; z <= maxMonoOrder_ - x - y; z++) {
	  xMonoOrder_.push_back(x);
	  yMonoOrder_.push_back(y);
	  zMonoOrder_.push_back(z);
	}
    break;
  case SPECIAL_POLY:
    xMonoOrder_.resize(7);
    yMonoOrder_.resize(7);
    zMonoOrder_.resize(7);
    maxPolyOrder_ = maxMonoOrder_;
    // Hard coded monomials in the following order:
    // 1, x, x^2, y, y^2, z, z^2
    xMonoOrder_[0]=0; yMonoOrder_[0]=0; zMonoOrder_[0]=0;
    xMonoOrder_[1]=1; yMonoOrder_[1]=0; zMonoOrder_[1]=0;
    xMonoOrder_[2]=2; yMonoOrder_[2]=0; zMonoOrder_[2]=0;
    xMonoOrder_[3]=0; yMonoOrder_[3]=1; zMonoOrder_[3]=0;
    xMonoOrder_[4]=0; yMonoOrder_[4]=2; zMonoOrder_[4]=0;
    xMonoOrder_[5]=0; yMonoOrder_[5]=0; zMonoOrder_[5]=1;
    xMonoOrder_[6]=0; yMonoOrder_[6]=0; zMonoOrder_[6]=2;
    break;
  default:
    throw domain_error("stencil.cc:setupPolyOrder-- unknown basis type!");
    break;
  }

  if (static_cast<int>(xMonoOrder_.size()) != numBasis_)
    throw domain_error("stencil.cc:setupPolyOrder-- inconsistent numBasis and polyTermIndex!");    
}

/**********************************************************************
 * setupBasisValue --
 **********************************************************************/
void
Stencil::setupBasisValue (
			  void)
{
  double e = 0.;
  vector<double> foo(numBasis_, e);
  basisValue_ = vector<vector<double> > (totalNumPoint_, foo);

  /* Testing in Matlab shows that the matrix 
   * starts to lose rank (numerically, not analytically) for dx<0.005.
   * At this point the result of the present computations would 
   * contain errors of RELATIVE magnitude one. The same problem occurs 
   * when dx becomes large. Obviously, the problem is a poor scaling 
   * of the columns of the "values" matrix. As a consequence a scaling 
   * of the values matrix is needed to ensure that the results are 
   * decent for all values of dx.
   */
  polyScale_ = 1./step_;
  vector<double> poly(numBasis_);
  for (int pointIndex = 0; pointIndex < totalNumPoint_; pointIndex++) {
    double x = points_[pointIndex].x() * step_;
    double y = points_[pointIndex].y() * step_;
    double z = points_[pointIndex].z() * step_;
    evaluateBasisFuncAtOnePoint(x, y, z, poly);
    for (int polyTermIndex = 0; polyTermIndex < numBasis_; polyTermIndex++) {
      basisValue_[pointIndex][polyTermIndex] = poly[polyTermIndex] * equaScale_[pointIndex];
    }
  }
}

/**********************************************************************
 * evaluateBasisFuncAtOnePoint --
 * This function is used by Stencil only.
 **********************************************************************/
void
Stencil::evaluateBasisFuncAtOnePoint (
				      double x,
				      double y,
				      double z,
				      vector<double>& funcValue) const
{
  x *= polyScale_;
  y *= polyScale_;
  z *= polyScale_;
  for (int polyTermIndex = 0; polyTermIndex < numBasis_; polyTermIndex++) {
    funcValue[polyTermIndex] = 
      intpow(x, xMonoOrder_[polyTermIndex]) *
      intpow(y, yMonoOrder_[polyTermIndex]) *
      intpow(z, zMonoOrder_[polyTermIndex]);
  }
}

/**********************************************************************
 * evaluateBasisFuncAtOnePoint --
 * This function is used by interpMat.
 **********************************************************************/
void
Stencil::evaluateBasisFuncAtOnePoint (
				      point3D& point,
				      const vector3D<double>& normal,
				      vector<double>& funcValue,
				      const DifferentialOperator operatorType) const
{
  point *= polyScale_;
  for (int polyTermIndex = 0; polyTermIndex < numBasis_; polyTermIndex++) {
    switch (operatorType) {
    case NONE:
      funcValue[polyTermIndex] = basisFuncValue(point, polyTermIndex);
      break;
    case D_DN:
      funcValue[polyTermIndex] = 
	normal.x() * basisFuncValue_dx(point, polyTermIndex) + 
	normal.y() * basisFuncValue_dy(point, polyTermIndex) + 
	normal.z() * basisFuncValue_dz(point, polyTermIndex);
      break;
    case D_DX:
      funcValue[polyTermIndex] = basisFuncValue_dx(point, polyTermIndex);
      break;
    case D_DY:
      funcValue[polyTermIndex] = basisFuncValue_dy(point, polyTermIndex);
      break;
    case D_DZ:
      funcValue[polyTermIndex] = basisFuncValue_dz(point, polyTermIndex);
      break;
    default:
      throw domain_error("unknown stencil basis type");
    }
  }
}

/**********************************************************************
 * compCoef --
 * This function is called by both interpMat and projectMat.
 * It computes a vector matrix product, vec * weight_
 * where vec contains either the value of many ploynomial terms evaluated
 * at collocation point (for the case of interpMat), or the value of
 * those polynomila terms integrated over the element (for the case of 
 * projectMat).
 * I don't want to reveal the internal data weight_. So I add this 
 * function to stencil class.
 **********************************************************************/
void
Stencil::compCoef (
		   const vector<double>& vec,
		   vector<double>& coef) const
{
  for (int col = 0; col < totalNumPoint_; col++) {
    coef[col] = 0.;
    for (int row = 0; row < numBasis_; row++) {
      coef[col] += vec[row] * weight_[row][col];
    }
  }
}


/**********************************************************************
 * setupWeight --
 **********************************************************************/
void
Stencil::setupWeight (
		      void)
{
  int size = max(numBasis_, totalNumPoint_);

  double **tmp;
  tmp = new double*[size];
  for (int i=0; i<size; i++) tmp[i] = new double[size];

  for (int pointIndex = 0; pointIndex < totalNumPoint_; pointIndex++) 
    for (int polyTermIndex = 0; polyTermIndex < numBasis_; polyTermIndex++) 
      tmp[pointIndex][polyTermIndex] = basisValue_[pointIndex][polyTermIndex];
  pseudo_inverse(tmp, totalNumPoint_, numBasis_);

  double e;
  vector<double> foo(totalNumPoint_, e);
  weight_ = vector<vector<double> > (numBasis_, foo);
  // rescale
  for (int polyTermIndex = 0; polyTermIndex < numBasis_; polyTermIndex++) 
    for (int pointIndex = 0; pointIndex < totalNumPoint_; pointIndex++) 
      weight_[polyTermIndex][pointIndex] = 
	tmp[polyTermIndex][pointIndex] * equaScale_[pointIndex];	

  // turn it to non-scale state
  for (int i = 0; i < totalNumPoint_; i++) equaScale_[i] = 1.;

  for (int i=0; i<size; i++) delete [] tmp[i];
  delete [] tmp;
}

/**********************************************************************
 * integrateBasisFuncOverElement --
 * moments: integral of basis Functions, the monomials, over an element
 **********************************************************************/
void
Stencil::integrateBasisFuncOverElement (
					const element& ele,
					const point3D& origin,
					const vector3D<double>& normal,
					vector<double>& moments,
					const DifferentialOperator operatorType)
{
  // I will add to this vector. So initializing to zero is necessary
  for (size_t i=0; i < moments.size(); i++)
    moments[i] = 0.;

  // find cubature points and weight in Berycentric coordinate system;
  if (! cubature_.setupComplete()) {
    int degree = maxPolyOrder_ + 1;
    int rule = 1;
    cubature_ = Cubature(degree, rule, Cubature::TRIANGLE);
  }

  // divide the element into many triangles because it is easier to use cubature
  vector<point3D> vertex(3);
  // adjust element verteices to the new origin
  // and scale the coordinate in accordance with similar precedure in 
  // computing weight_
  vertex[0] = (ele.vertex(0) - origin) * polyScale_;
  for (size_t i=1; i < ele.shape() - 1; i++) {
    vertex[1] = (ele.vertex(i) - origin) * polyScale_;
    vertex[2] = (ele.vertex(i+1) - origin) * polyScale_;
    addIntegralOverTriangle(vertex, normal, moments, operatorType);
  }

}

/**********************************************************************
 * addIntegralOverTriangle --
 **********************************************************************/
void
Stencil::addIntegralOverTriangle (
				  const vector<point3D>& vertex, 
				  const vector3D<double>& normal,
				  vector<double>& moments,
				  const DifferentialOperator operatorType) const
{
  double area = compTriangleArea(vertex[0], vertex[1], vertex[2])
    / (0.5 * polyScale_ * polyScale_);

  for (int pointIndex = 0; pointIndex < cubature_.numPoint() ; pointIndex++) {
    point3D point = cubature_.point(vertex, pointIndex);
    double weight = cubature_.weight(area, pointIndex);
    for (int polyTermIndex = 0; polyTermIndex < numBasis_; polyTermIndex++) {

      switch (operatorType) {
      case NONE:
	moments[polyTermIndex] += basisFuncValue(point, polyTermIndex) * weight;
	break;
      case D_DN:
	moments[polyTermIndex] += 
	  (normal.x() * basisFuncValue_dx(point, polyTermIndex) + 
	   normal.y() * basisFuncValue_dy(point, polyTermIndex) + 
	   normal.z() * basisFuncValue_dz(point, polyTermIndex) ) 
	   * weight;
	break;
      case D_DX:
	moments[polyTermIndex] += basisFuncValue_dx(point, polyTermIndex) * weight;
	break;
      case D_DY:
	moments[polyTermIndex] += basisFuncValue_dy(point, polyTermIndex) * weight;
	break;
      case D_DZ:
	moments[polyTermIndex] += basisFuncValue_dz(point, polyTermIndex) * weight;
      break;
      default:
	throw domain_error("unknown stencil basis type");
      }
    }
  }
}

/**********************************************************************
 * compTriangleArea --
 **********************************************************************/
inline double
Stencil::compTriangleArea (
			   const point3D& v1,
			   const point3D& v2,
			   const point3D& v3) const 
{
  vector3D<double> edge1 = v2 - v1;
  vector3D<double> edge2 = v2 - v3;
  vector3D<double> vec = crossProd(edge1, edge2);
  return 0.5 * length(vec);
}

/**********************************************************************
 * basisFuncValue --
 **********************************************************************/
inline double
Stencil::basisFuncValue (
			 const point3D& point,
			 const int polyTermIndex) const
{
  return 
    intpow(point.x(), xMonoOrder_[polyTermIndex]) *
    intpow(point.y(), yMonoOrder_[polyTermIndex]) *
    intpow(point.z(), zMonoOrder_[polyTermIndex]);

}

/**********************************************************************
 * basisFuncValue_dx --
 **********************************************************************/
inline double
Stencil::basisFuncValue_dx (
			    const point3D& point,
			    const int polyTermIndex) const
{
  if (xMonoOrder_[polyTermIndex] == 0) {
    return 0.;
  } else {
    return
      xMonoOrder_[polyTermIndex] * intpow(point.x(), xMonoOrder_[polyTermIndex]-1) *
      intpow(point.y(), yMonoOrder_[polyTermIndex]) *
      intpow(point.z(), zMonoOrder_[polyTermIndex]) * polyScale_;
  }
}

/**********************************************************************
 * basisFuncValue_dy --
 **********************************************************************/
inline double
Stencil::basisFuncValue_dy (
			    const point3D& point,
			    const int polyTermIndex) const
{
  if (yMonoOrder_[polyTermIndex] == 0) {
    return 0.;
  } else {
    return
      intpow(point.x(), xMonoOrder_[polyTermIndex]) *
      yMonoOrder_[polyTermIndex] * intpow(point.y(), yMonoOrder_[polyTermIndex]-1) *
      intpow(point.z(), zMonoOrder_[polyTermIndex]) * polyScale_;
  }
}

/**********************************************************************
 * basisFuncValue_dz --
 **********************************************************************/
inline double
Stencil::basisFuncValue_dz (
			    const point3D& point,
			    const int polyTermIndex) const
{
  if (zMonoOrder_[polyTermIndex] == 0) {
    return 0.;
  } else {
    return
      intpow(point.x(), xMonoOrder_[polyTermIndex]) *
      intpow(point.y(), yMonoOrder_[polyTermIndex]) *
      zMonoOrder_[polyTermIndex] * intpow(point.z(), zMonoOrder_[polyTermIndex]-1) * polyScale_;
  }
}

/**********************************************************************
 * intpow --
 * By Dennis Wu
 **********************************************************************/
inline double 
Stencil::intpow(
		const double d, 
		int e) const 
{
  double product = 1.;
  double increment = d;
  
  while(e>0){
    if ((e%2)!=0){
      product = product * increment;
      e=e-1;
    }else{
      increment=increment*increment;
      e=e/2;
    }
  }
  return product;
}
